
debugger;

main();

function main() {
  const canvas = document.querySelector('#glcanvas');
  gl = canvas.getContext('webgl2');

  if (!gl) {
    alert('Unable to initialize WebGL. Your browser or machine may not support it.');
    return;
  }

  const vsSource = `#version 300 es
    in vec4 aVertexPosition;
    in vec4 aVertexColor;

    uniform mat4 uModelViewMatrix;
    uniform mat4 uProjectionMatrix;

    out lowp vec4 vColor;

    void main(void) {
      gl_Position = uProjectionMatrix * uModelViewMatrix * aVertexPosition;
      vColor = aVertexColor;
    }
  `;

  const fsSource = `#version 300 es
    in lowp vec4 vColor;
    out lowp vec4 fragColor;

    void main(void) {
      fragColor = vColor;
    }
  `;

  const shaderProgram = initShaderProgram(gl, vsSource, fsSource);

  programInfo = {
    program: shaderProgram,
    attribLocations: {
      vertexPosition: gl.getAttribLocation(shaderProgram, 'aVertexPosition'),
      vertexColor: gl.getAttribLocation(shaderProgram, 'aVertexColor'),
    },
    uniformLocations: {
      projectionMatrix: gl.getUniformLocation(shaderProgram, 'uProjectionMatrix'),
      modelViewMatrix: gl.getUniformLocation(shaderProgram, 'uModelViewMatrix'),
    },
  };

var tfShaderSourcePS = `#version 300 es
  out lowp vec4 fragColor;
  void main()
  {
      fragColor = vec4(0,0,0,0);
  }
  `;

var tfShaderSourceVS = `#version 300 es
  in vec2 inPos;
  in vec2 prevPos;
  in vec2 prevPrevPos;

  uniform int havePrev;
  uniform int havePrevPrev;

  out vec2 outPos;

  void main()
  {
      if (havePrev == 1)
      {
          vec2 dm = vec2(0.0, 0.0);

          if (havePrevPrev == 1)
          {
              dm = prevPos - prevPrevPos;
          }

          vec2 delta = inPos;
          outPos = prevPos + dm + delta;
      }

      else
      {
          outPos = inPos;
      }
  }
  `;

  transformFeedbackProgram = initShaderProgram(gl, tfShaderSourceVS, tfShaderSourcePS, function(program) {
    gl.transformFeedbackVaryings(program, ["outPos"], gl.SEPARATE_ATTRIBS);
  });

  transformFeedbackProgramInfo = {
    program: transformFeedbackProgram,
    attribLocations: {
      input: gl.getAttribLocation(transformFeedbackProgram, 'inPos'),
      prev: gl.getAttribLocation(transformFeedbackProgram, 'prevPos'),
      prevPrev: gl.getAttribLocation(transformFeedbackProgram, 'prevPrevPos'),
    },
    uniformLocations: {
      havePrev: gl.getUniformLocation(transformFeedbackProgram, 'havePrev'),
      havePrevPrev: gl.getUniformLocation(transformFeedbackProgram, 'havePrevPrev'),
    }
  };

  buffers = initBuffers(gl);  

  for (var i = 0 ; i < 3 ; ++i) {
    gl.bindTransformFeedback(gl.TRANSFORM_FEEDBACK, buffers.transformFeedbacks[i]);
    gl.bindBufferBase(gl.TRANSFORM_FEEDBACK_BUFFER, 0, buffers.outputBuffers[i]);
    gl.bufferData(gl.TRANSFORM_FEEDBACK_BUFFER, 4 * 2 * 4, gl.STATIC_DRAW);
  }

  gl.bindTransformFeedback(gl.TRANSFORM_FEEDBACK, null);
  
  frame = 0;

  drawScene();
}

function initBuffers(gl) {

  const positions = [
     1.0,  1.0,
    -1.0,  1.0,
     1.0, -1.0,
    -1.0, -1.0,
  ];

  const deltas1 = [
    0.5, 0.0,
    0.5, 0.0,
    0.5, 0.0,
    0.5, 0.0,
  ];

  const deltas2 = [
    0.0, 0.5,
    0.0, 0.5,
    0.0, 0.5,
    0.0, 0.5,
  ];

  var colors = [
    1.0,  1.0,  1.0,  1.0,    // white
    1.0,  0.0,  0.0,  1.0,    // red
    0.0,  1.0,  0.0,  1.0,    // green
    0.0,  0.0,  1.0,  1.0,    // blue
  ];

  const colorBuffer = gl.createBuffer();
  gl.bindBuffer(gl.ARRAY_BUFFER, colorBuffer);
  gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(colors), gl.STATIC_DRAW);

  return {
    staging: gl.createBuffer(),
    outputBuffers: [gl.createBuffer(), gl.createBuffer(), gl.createBuffer()],
    transformFeedbacks: [gl.createTransformFeedback(), gl.createTransformFeedback(), gl.createTransformFeedback()],
    color: colorBuffer,
    positions:positions,
    deltas1:deltas1,
    deltas2:deltas2,
  };
}

function drawScene() {
  
  gl.useProgram(transformFeedbackProgramInfo.program);
  gl.bindBuffer(gl.ARRAY_BUFFER, null);

  if (frame == 0) {

    gl.bindBuffer(gl.ARRAY_BUFFER, buffers.staging);
    gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(buffers.positions), gl.STATIC_DRAW);
    gl.enableVertexAttribArray(transformFeedbackProgramInfo.input);
    gl.vertexAttribPointer(transformFeedbackProgramInfo.attribLocations.input, 2, gl.FLOAT, false, 0, 0);

    gl.disableVertexAttribArray(transformFeedbackProgramInfo.attribLocations.prev);
    gl.disableVertexAttribArray(transformFeedbackProgramInfo.attribLocations.prevPrev);

    gl.uniform1i(transformFeedbackProgramInfo.uniformLocations.havePrev, 0);
    gl.uniform1i(transformFeedbackProgramInfo.uniformLocations.havePrevPrev, 0);
  }

  else if (frame == 1) {
    
    gl.bindBuffer(gl.ARRAY_BUFFER, buffers.staging);
    gl.bufferData(gl.ARRAY_BUFFER,  new Float32Array(buffers.deltas1), gl.STATIC_DRAW);
    gl.enableVertexAttribArray(transformFeedbackProgramInfo.input);

    gl.vertexAttribPointer(transformFeedbackProgramInfo.input, 2, gl.FLOAT, false, 0, 0);

    gl.bindBuffer(gl.ARRAY_BUFFER, buffers.outputBuffers[0]);
    gl.enableVertexAttribArray(transformFeedbackProgramInfo.prev);
    gl.vertexAttribPointer(transformFeedbackProgramInfo.attribLocations.prev, 2, gl.FLOAT, false, 0, 0);

    gl.disableVertexAttribArray(transformFeedbackProgramInfo.attribLocations.prevPrev);

    gl.uniform1i(transformFeedbackProgramInfo.uniformLocations.havePrev, 1);
    gl.uniform1i(transformFeedbackProgramInfo.uniformLocations.havePrevPrev, 0);
  }

  else if (frame == 2) {

    gl.bindBuffer(gl.ARRAY_BUFFER, buffers.staging);
    gl.bufferData(gl.ARRAY_BUFFER,  new Float32Array(buffers.deltas2), gl.STATIC_DRAW);
    gl.enableVertexAttribArray(transformFeedbackProgramInfo.input);

    gl.vertexAttribPointer(transformFeedbackProgramInfo.attribLocations.input, 2, gl.FLOAT, false, 0, 0);

    gl.bindBuffer(gl.ARRAY_BUFFER, buffers.outputBuffers[1]);
    gl.enableVertexAttribArray(transformFeedbackProgramInfo.attribLocations.prev);
    gl.vertexAttribPointer(transformFeedbackProgramInfo.attribLocations.prev, 2, gl.FLOAT, false, 0, 0);

    gl.bindBuffer(gl.ARRAY_BUFFER, buffers.outputBuffers[0]);
    gl.enableVertexAttribArray(transformFeedbackProgramInfo.attribLocations.prevPrev);
    gl.vertexAttribPointer(transformFeedbackProgramInfo.attribLocations.prevPrev, 2, gl.FLOAT, false, 0,0);

    gl.uniform1i(transformFeedbackProgramInfo.uniformLocations.havePrev, 1);
    gl.uniform1i(transformFeedbackProgramInfo.uniformLocations.havePrevPrev, 1);
  }

  gl.bindTransformFeedback(gl.TRANSFORM_FEEDBACK, buffers.transformFeedbacks[frame]);
  gl.enable(gl.RASTERIZER_DISCARD);
  gl.beginTransformFeedback(gl.POINTS);
  gl.drawArrays(gl.POINTS, 0, 4);
  gl.endTransformFeedback();
  gl.disable(gl.RASTERIZER_DISCARD);
  gl.bindTransformFeedback(gl.TRANSFORM_FEEDBACK, null);

  // for input to real draw below
  gl.bindBuffer(gl.ARRAY_BUFFER, buffers.outputBuffers[frame]);
  
  frame = (frame + 1 ) % 3;

  gl.clearColor(0.0, 0.0, 0.0, 1.0);
  gl.clearDepth(1.0);
  gl.enable(gl.DEPTH_TEST);
  gl.depthFunc(gl.LEQUAL);

  gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);

  const fieldOfView = 45 * Math.PI / 180;   // in radians
  const aspect = gl.canvas.clientWidth / gl.canvas.clientHeight;
  const zNear = 0.1;
  const zFar = 100.0;
  const projectionMatrix = mat4.create();

  mat4.perspective(projectionMatrix,
                   fieldOfView,
                   aspect,
                   zNear,
                   zFar);

  const modelViewMatrix = mat4.create();

  mat4.translate(modelViewMatrix,     // destination matrix
                 modelViewMatrix,     // matrix to translate
                 [-0.0, 0.0, -6.0]);  // amount to translate

  {
    const numComponents = 2;
    const type = gl.FLOAT;
    const normalize = false;
    const stride = 0;
    const offset = 0;
    //gl.bindBuffer(gl.ARRAY_BUFFER, buffers.position);
    gl.vertexAttribPointer(
        programInfo.attribLocations.vertexPosition,
        numComponents,
        type,
        normalize,
        stride,
        offset);
    gl.enableVertexAttribArray(
        programInfo.attribLocations.vertexPosition);
  }

  {
    const numComponents = 4;
    const type = gl.FLOAT;
    const normalize = false;
    const stride = 0;
    const offset = 0;
    gl.bindBuffer(gl.ARRAY_BUFFER, buffers.color);
    gl.vertexAttribPointer(
        programInfo.attribLocations.vertexColor,
        numComponents,
        type,
        normalize,
        stride,
        offset);
    gl.enableVertexAttribArray(
        programInfo.attribLocations.vertexColor);
  }

  gl.useProgram(programInfo.program);

  gl.uniformMatrix4fv(
      programInfo.uniformLocations.projectionMatrix,
      false,
      projectionMatrix);
  gl.uniformMatrix4fv(
      programInfo.uniformLocations.modelViewMatrix,
      false,
      modelViewMatrix);

  {
    const offset = 0;
    const vertexCount = 4;
    gl.drawArrays(gl.TRIANGLE_STRIP, offset, vertexCount);
  }

  requestAnimationFrame(drawScene);
}

function initShaderProgram(gl, vsSource, fsSource, prelink) {
  const vertexShader = loadShader(gl, gl.VERTEX_SHADER, vsSource);
  const fragmentShader = loadShader(gl, gl.FRAGMENT_SHADER, fsSource);

  const shaderProgram = gl.createProgram();
  gl.attachShader(shaderProgram, vertexShader);
  gl.attachShader(shaderProgram, fragmentShader);

  if (prelink) {
    prelink(shaderProgram);
  }

  gl.linkProgram(shaderProgram);

  if (!gl.getProgramParameter(shaderProgram, gl.LINK_STATUS)) {
    alert('Unable to initialize the shader program: ' + gl.getProgramInfoLog(shaderProgram));
    return null;
  }

  return shaderProgram;
}

function loadShader(gl, type, source) {
  const shader = gl.createShader(type);

  gl.shaderSource(shader, source);

  gl.compileShader(shader);

  if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
    alert('An error occurred compiling the shaders: ' + gl.getShaderInfoLog(shader));
    gl.deleteShader(shader);
    return null;
  }

  return shader;
}

